@@ -7,13 +7,13 @@ from doozerlib import version
 from doozerlib import Runtime, Dir
 from doozerlib import state
 from doozerlib.model import Missing
-from doozerlib.brew import get_watch_task_info_copy
+from doozerlib.brew import get_watch_task_info_copy, get_latest_builds, list_archives_by_builds
 from doozerlib import metadata
 from doozerlib.config import MetaDataConfig as mdc
 from doozerlib.cli import cli_opts
 from doozerlib.exceptions import DoozerFatalError
 from doozerlib import exectools
-from doozerlib.util import green_prefix, red_prefix, green_print, red_print, yellow_print, yellow_prefix, color_print, dict_get
+from doozerlib.util import green_prefix, red_prefix, green_print, red_print, yellow_print, yellow_prefix, color_print, dict_get, analyze_debug_timing
 from doozerlib import operator_metadata
 import click
 import os
@@ -27,6 +27,7 @@ import tempfile
 import traceback
 import koji
 import io
+import datetime
 from numbers import Number
 from multiprocessing.pool import ThreadPool
 from multiprocessing import cpu_count
@@ -128,6 +129,8 @@ def print_version(ctx, param, value):
               help="Path to rhpkg config file to use instead of system default")
 @click.option("--brew-tag", metavar="BREW_TAG",
               help="Override brew tag to expect for images built")
+@click.option("--cache-dir", metavar="DIR", required=False, default=None,
+              help="A directory in which reference git repos can be stored for caching purposes")
 @click.option("--profile", metavar="NAME", default="", help="Name of build profile")
 @click.pass_context
 def cli(ctx, **kwargs):
@@ -186,11 +189,9 @@ def cli(ctx, **kwargs):
     return ctx
 
 
-def validate_semver_major_minor_patch(ctx, param, version):
+def validate_version(ctx, param, version):
     """
-    For non-None, non-auto values, ensures that the incoming parameter meets the criteria vX.Y.Z or X.Y.Z.
-    If minor or patch is not supplied, the value is modified to possess
-    minor.major to meet semver requirements.
+    For non-None, non-auto values, ensures that first 3 version segments are integers (with optional v prefix)
     :param ctx: Click context
     :param param: The parameter specified on the command line
     :param version: The version specified on the command line
@@ -202,15 +203,12 @@ def validate_semver_major_minor_patch(ctx, param, version):
     vsplit = version.split(".")
     try:
         int(vsplit[0].lstrip('v'))
-        minor_version = int('0' if len(vsplit) < 2 else vsplit[1])
-        patch_version = int('0' if len(vsplit) < 3 else vsplit[2])
+        int('0' if len(vsplit) < 2 else vsplit[1])
+        int('0' if len(vsplit) < 3 else vsplit[2])
     except ValueError:
         raise click.BadParameter('Expected integers in version fields')
 
-    if len(vsplit) > 3:
-        raise click.BadParameter('Expected X, X.Y, or X.Y.Z (with optional "v" prefix)')
-
-    return f'{vsplit[0]}.{minor_version}.{patch_version}'
+    return version
 
 
 option_commit_message = click.option("--message", "-m", cls=RemoteRequired, metavar='MSG', help="Commit message for dist-git.")
@@ -254,7 +252,7 @@ def rpms_clone_sources(runtime, output_yml):
 
 
 @cli.command("rpms:build", help="Build rpms in the group or given by --rpms.")
-@click.option("--version", metavar='VERSION', default=None, callback=validate_semver_major_minor_patch,
+@click.option("--version", metavar='VERSION', default=None, callback=validate_version,
               help="Version string to populate in specfile.", required=True)
 @click.option("--release", metavar='RELEASE', default=None,
               help="Release label to populate in specfile.", required=True)
@@ -317,7 +315,7 @@ def images_push_distgit(runtime):
 @cli.command("images:update-dockerfile", short_help="Update a group's distgit Dockerfile from metadata.")
 @click.option("--stream", metavar="ALIAS REPO/NAME:TAG", nargs=2, multiple=True,
               help="Associate an image name with a given stream alias.  [multiple]")
-@click.option("--version", metavar='VERSION', default=None, callback=validate_semver_major_minor_patch,
+@click.option("--version", metavar='VERSION', default=None, callback=validate_version,
               help="Version string to populate in Dockerfiles. \"auto\" gets version from atomic-openshift RPM")
 @click.option("--release", metavar='RELEASE', default=None,
               help="Release label to populate in Dockerfiles (or + to bump).")
@@ -346,7 +344,7 @@ def images_update_dockerfile(runtime, stream, version, release, repo_type, messa
         yellow_print('images:update-dockerfile is not valid when using --local, use images:rebase instead')
         sys.exit(1)
 
-    runtime.initialize(validate_content_sets=True)
+    runtime.initialize(validate_content_sets=True, clone_distgits=True)
 
     # This is ok to run if automation is frozen as long as you are not pushing
     if push:
@@ -374,37 +372,31 @@ def images_update_dockerfile(runtime, stream, version, release, repo_type, messa
             "invalid version string: {}, expecting like v3.4 or v1.2.3".format(version)
         )
 
-    runtime.clone_distgits()
     metas = runtime.image_metas()
     lstate['total'] = len(metas)
 
-    for image in metas:
+    def dgr_update(image_meta):
         try:
-            dgr = image.distgit_repo()
+            dgr = image_meta.distgit_repo()
             (real_version, real_release) = dgr.update_distgit_dir(version, release)
             dgr.commit(message)
             dgr.tag(real_version, real_release)
-            state.record_image_success(lstate, image)
+            state.record_image_success(lstate, image_meta)
+            if push:
+                (meta, success) = dgr.push()
+                if success is not True:
+                    state.record_image_fail(lstate, meta, success)
         except Exception as ex:
-            msg = None
-            try:
-                msg = ex.strerror
-            except:
-                msg = str(ex)
-
-            state.record_image_fail(lstate, image, msg, runtime.logger)
-
-    try:
-        if push:
-            res = runtime.push_distgits()
-
-            for name, r in res:
-                if r is not True:
-                    state.record_image_fail(lstate, name, r)
-    except Exception as ex:
-        lstate['status'] = state.STATE_FAIL
-        raise DoozerFatalError(getattr(ex, 'message', repr(ex)))
+            msg = str(ex)
+            state.record_image_fail(lstate, image_meta, msg, runtime.logger)
+            return False
+        return True
 
+    jobs = runtime.parallel_exec(
+        lambda image_meta, terminate_event: dgr_update(image_meta),
+        metas,
+    )
+    jobs.get()
     state.record_image_finish(lstate)
 
     failed = []
@@ -458,7 +450,7 @@ def config_scan_source_changes(runtime, as_yaml):
 @cli.command("images:rebase", short_help="Refresh a group's distgit content from source content.")
 @click.option("--stream", metavar="ALIAS REPO/NAME:TAG", nargs=2, multiple=True,
               help="Associate an image name with a given stream alias.  [multiple]")
-@click.option("--version", metavar='VERSION', default=None, callback=validate_semver_major_minor_patch,
+@click.option("--version", metavar='VERSION', default=None, callback=validate_version,
               help="Version string to populate in Dockerfiles. \"auto\" gets version from atomic-openshift RPM")
 @click.option("--release", metavar='RELEASE', default=None, help="Release string to populate in Dockerfiles.")
 @click.option("--repo-type", metavar="REPO_TYPE", envvar="OIT_IMAGES_REPO_TYPE",
@@ -514,51 +506,49 @@ def images_rebase(runtime, stream, version, release, repo_type, message, push):
     runtime.clone_distgits()
     metas = runtime.image_metas()
     lstate['total'] = len(metas)
-    for image in metas:
+
+    def dgr_rebase(image_meta):
         try:
-            dgr = image.distgit_repo()
+            dgr = image_meta.distgit_repo()
             (real_version, real_release) = dgr.rebase_dir(version, release)
             sha = dgr.commit(message, log_diff=True)
             dgr.tag(real_version, real_release)
             runtime.add_record(
                 "distgit_commit",
-                distgit=image.qualified_name,
-                image=image.config.name,
+                distgit=image_meta.qualified_name,
+                image=image_meta.config.name,
                 sha=sha,
             )
-            state.record_image_success(lstate, image)
+            state.record_image_success(lstate, image_meta)
+
+            if push:
+                (meta, success) = dgr.push()
+                if success is not True:
+                    state.record_image_fail(lstate, meta, success)
+
         except Exception as ex:
             # Only the message will recorded in the state. Make sure we print out a stacktrace in the logs.
             traceback.print_exc()
 
-            owners = image.config.owners
+            owners = image_meta.config.owners
             owners = ",".join(list(owners) if owners is not Missing else [])
             runtime.add_record(
                 "distgit_commit_failure",
-                distgit=image.qualified_name,
-                image=image.config.name,
+                distgit=image_meta.qualified_name,
+                image=image_meta.config.name,
                 owners=owners,
                 message=str(ex).replace("|", ""),
             )
+            msg = str(ex)
+            state.record_image_fail(lstate, image_meta, msg, runtime.logger)
+            return False
+        return True
 
-            msg = None
-            try:
-                msg = ex.strerror
-            except:
-                msg = str(ex)
-
-            state.record_image_fail(lstate, image, msg, runtime.logger)
-
-    try:
-        if push:
-            res = runtime.push_distgits()
-
-            for name, r in res:
-                if r is not True:
-                    state.record_image_fail(lstate, name, r)
-    except Exception as ex:
-        lstate['status'] = state.STATE_FAIL
-        raise DoozerFatalError(repr(ex))
+    jobs = runtime.parallel_exec(
+        lambda image_meta, terminate_event: dgr_rebase(image_meta),
+        metas,
+    )
+    jobs.get()
 
     state.record_image_finish(lstate)
 
@@ -1710,65 +1700,55 @@ that particular tag.
     # This will map[arch] -> map[image_name] -> { version: version, release: release, image_src: image_src }
     mirroring = {}
 
+    ose_prefixed_images = []
     for image in images:
-        try:
-            # Per clayton:
-            """Tim Bielawa: note to self: is only for `ose-` prefixed images
-            Clayton Coleman: Yes, Get with the naming system or get out of town
-            """
-            if 'ose' not in image.image_name_short:
-                invalid_name_items.append(image.image_name_short)
-                red_print("NOT adding to IS (does not meet name/version conventions): {}".format(image.image_name_short))
-                continue
+        # Per clayton:
+        """Tim Bielawa: note to self: is only for `ose-` prefixed images
+        Clayton Coleman: Yes, Get with the naming system or get out of town
+        """
+        if not image.image_name_short.startswith("ose-"):
+            invalid_name_items.append(image.image_name_short)
+            red_print("NOT adding to IS (does not meet name/version conventions): {}".format(image.image_name_short))
+            continue
+        ose_prefixed_images.append(image)
 
-            try:
-                _, version, release = image.get_latest_build_info()
-                src = 'registry-proxy.engineering.redhat.com/rh-osbs/openshift-{}:{}-{}'.format(
-                    image.image_name_short, version, release)
-                click.echo('Processing: {}'.format(src))
-            except IOError:
-                yellow_print('Unable to find build for: {}'.format(image.image_name_short))
-                no_build_items.append(image.image_name_short)
-                continue
+    runtime.logger.info("Fetching latest image builds from Brew...")
+    tag_component_tuples = [(image.brew_tag(), image.get_component_name()) for image in ose_prefixed_images]
+    brew_session = koji.ClientSession(runtime.group_config.urls.brewhub)
+    latest_builds = get_latest_builds(tag_component_tuples, brew_session)
 
-            # The tag that will be used in the imagestreams
-            tag_name = image.image_name_short.replace('ose-', '')
-
-            src_repo = src.split(':')[0].split('@')[0]  # strip off image tag or sha to just get base repo
-            # skopeo will print out json data for the image.
-            manifest_info_str = subprocess.check_output(["skopeo", "inspect", "-raw", 'docker://{}'.format(src)], stderr=subprocess.STDOUT)
-            manifest_info = json.loads(manifest_info_str)
-
-            mediaType = manifest_info.get('mediaType', None)
-            if not mediaType:
-                raise DoozerFatalError('Unable to inspect {} for mediaType:\n {}'.format(src, manifest_info_str))
-
-            def add_tag_image(arch, arch_image):
-                # mirroring[arch][tag_name] = image_url
-                mirroring[arch] = mirroring.get(arch, {})
-                mirroring[arch][tag_name] = {'version': version, 'release': release, 'image_src': arch_image}
-
-            # A manifest.list is a list of images for different arches
-            if mediaType == "application/vnd.docker.distribution.manifest.list.v2+json":
-                for arch_manifest in manifest_info['manifests']:
-                    arch = arch_manifest['platform']['architecture']
-                    if arch == 'amd64':
-                        arch = 'x86_64'
-                    arch_image = '{}@{}'.format(src_repo, arch_manifest['digest'])
-                    add_tag_image(arch, arch_image)
-            else:
-                # If this is not a manifest list, assume this is x86_64
-                arch = 'x86_64'
-                arch_image = src
-                add_tag_image(arch, arch_image)
+    runtime.logger.info("Fetching pullspecs...")
+    build_ids = [builds[0]["id"] if builds else 0 for builds in latest_builds]
+    archives_list = list_archives_by_builds(build_ids, "image", brew_session)
 
+    runtime.logger.info("Creating mirroring lists...")
+    for i, image in enumerate(ose_prefixed_images):
+        latest_build = latest_builds[i]
+        archives = archives_list[i]
+        error = None
+        if not (latest_build and archives):  # build or archive doesn't exist
+            error = f"Unable to find build for: {image.image_name_short}"
+            no_build_items.append(image.image_name_short)
+            state.record_image_fail(lstate, image, error, runtime.logger)
+        else:
+            for archive in archives:
+                arch = archive["arch"]
+                pullspecs = archive["extra"]["docker"]["repositories"]
+                if not pullspecs or ":" not in pullspecs[-1]:  # in case of no pullspecs or invalid format
+                    error = f"Unable to find pullspecs for: {image.image_name_short}"
+                    red_print(error, file=sys.stderr)
+                    state.record_image_fail(lstate, image, error, runtime.logger)
+                    break
+                # The tag that will be used in the imagestreams
+                tag_name = image.image_name_short
+                tag_name = tag_name[4:] if tag_name.startswith("ose-") else tag_name  # it _should_ but... to be safe
+                runtime.logger.info(f"Adding image {pullspecs[-1]} to the {arch} mirroring list with imagestream tag {tag_name}...")
+                mirroring.setdefault(arch, {})[tag_name] = {'version': latest_build[0]["version"], 'release': latest_build[0]["release"], 'image_src': pullspecs[-1]}
+        if not error:
             state.record_image_success(lstate, image)
-
-            # End 'for image in images' loop
-            #############################################################
-        except state.DoozerStateError as serr:
-            red_print(str(serr))
-            state.record_image_fail(lstate, image, str(serr), runtime.logger)
+        else:
+            red_print(error, file=sys.stderr)
+            state.record_image_fail(lstate, image, error, runtime.logger)
 
     for arch in mirroring:
 
@@ -1786,11 +1766,12 @@ that particular tag.
                 arch_ext = ''
             else:
                 arch_ext = '-{}'.format(arch)
-            return 'quay.io/{orgrepo}:{version}-{release}{arch_ext}-ose-{tag_name}'.format(orgrepo=orgrepo,
-                                                                                           version=entry['version'],
-                                                                                           release=entry['release'],
-                                                                                           arch_ext=arch_ext,
-                                                                                           tag_name=tag_name)
+            return 'quay.io/{orgrepo}:{version}-{release}{arch_ext}-ose-{tag_name}-{uuid}'.format(orgrepo=orgrepo,
+                                                                                                  version=entry['version'],
+                                                                                                  release=entry['release'],
+                                                                                                  arch_ext=arch_ext,
+                                                                                                  tag_name=tag_name,
+                                                                                                  uuid=runtime.uuid)
 
         # Save the default SRC=DEST 'oc image mirror' input to a file for
         # later.
@@ -2133,6 +2114,36 @@ def operator_metadata_latest_build(runtime, nvr, stream, operator_list):
             click.echo(operator_metadata.OperatorMetadataLatestNvrReporter(build, stream, runtime).get_latest_build())
 
 
+@cli.command("analyze:debug-log", short_help="Output an analysis of the debug log")
+@click.argument("debug_log", nargs=1)
+def analyze_debug_log(debug_log):
+    """
+    Specify a doozer working directory debug.log as the argument. A table with be printed to stdout.
+    The top row of the table will be all the threads that existed during the doozer runtime.
+    T1, T2, T3, ... each represent an independent thread which was created and performed some action.
+
+    The first column of the graph indicates a time interval. Each interval is 10 seconds. Actions
+    performed by each thread are grouped into the interval in which those actions occurred. The
+    action performed by a thread is aligned and printed under the thread's column.
+
+    Example 1:
+    *     T1    T2    T3
+    0     loading metadata A                          # During the first 10 seconds, T1 is perform actions.
+     0    loading metaddata B                         # Multiple events can happen during the same interval.
+     0    loading metaddata C
+
+    Example 2:
+    *     T1    T2    T3
+    0     loading metadata A                          # During the first 10 seconds, T1 is perform actions.
+     0    loading metaddata B                         # Multiple events can happen during the same interval.
+    1           cloning                               # During seconds 10-20, three threads start cloning.
+     1    cloning
+     1                cloning
+    """
+    f = os.path.abspath(debug_log)
+    analyze_debug_timing(f)
+
+
 def main():
     if sys.version_info.major < 3:
         yellow_print("DEPRECATION: Python 2.7 will reach the end of its life on January 1st, 2020. Please upgrade your Python as Python 2.7 won't be maintained after that date. A future version of Doozer will drop support for Python 2.7.",
