@@ -20,6 +20,7 @@ import urllib.parse
 import signal
 import io
 import pathlib
+import subprocess
 
 from doozerlib import gitdata
 from . import logutil
@@ -141,6 +142,8 @@ class Runtime(object):
         self._remove_tmp_working_dir = False
         self.group_config = None
 
+        self.cwd = os.getcwd()
+
         # If source needs to be cloned by oit directly, the directory in which it will be placed.
         self.sources_dir = None
 
@@ -310,6 +313,9 @@ class Runtime(object):
         if not os.path.isdir(self.flags_dir):
             os.mkdir(self.flags_dir)
 
+        if self.cache_dir:
+            self.cache_dir = os.path.abspath(self.cache_dir)
+
         self.group_dir = self.gitdata.data_dir
 
         # register the sources
@@ -739,7 +745,7 @@ class Runtime(object):
             replace_vars = self.group_config.vars.primitive()
         data_obj = self.gitdata.load_data(path='images', key=distgit_name, replace_vars=replace_vars)
         if not data_obj:
-            raise DoozerFatalError('Unable to resovle image metadata for {}'.format(distgit_name))
+            raise DoozerFatalError('Unable to resolve image metadata for {}'.format(distgit_name))
 
         meta = ImageMetadata(self, data_obj)
         if add:
@@ -794,6 +800,59 @@ class Runtime(object):
 
         return remote_https
 
+    def git_clone(self, remote_url, target_dir, gitargs=None, set_env=None, timeout=0):
+        gitargs = gitargs or []
+        set_env = set_env or []
+
+        if self.cache_dir:
+            git_cache_dir = os.path.join(self.cache_dir, self.user, 'git')
+            util.mkdirs(git_cache_dir)
+            normalized_url = util.convert_remote_git_to_https(remote_url)
+            # Strip special chars out of normalized url to create a human friendly, but unique filename
+            file_friendly_url = normalized_url.split('//')[-1].replace('/', '_')
+            repo_dir = os.path.join(git_cache_dir, file_friendly_url)
+            self.logger.info(f'Cache for {remote_url} going to f{repo_dir}')
+
+            if not os.path.exists(repo_dir):
+                self.logger.info(f'Initializing cache directory for git remote: {remote_url}')
+
+                # If the cache directory for this repo does not exist yet, we will create one.
+                # But we must do so carefully to minimize races with any other doozer instance
+                # running on the machine.
+                with self.get_dir_lock(repo_dir):  # also make sure we cooperate with other threads in this process.
+                    tmp_repo_dir = tempfile.mkdtemp(dir=git_cache_dir)
+                    exectools.cmd_assert(f'git init --bare {tmp_repo_dir}')
+                    with Dir(tmp_repo_dir):
+                        exectools.cmd_assert(f'git remote add origin {remote_url}')
+
+                    try:
+                        os.rename(tmp_repo_dir, repo_dir)
+                    except:
+                        # There are two categories of failure
+                        # 1. Another doozer instance already created the directory, in which case we are good to go.
+                        # 2. Something unexpected is preventing the rename.
+                        if not os.path.exists(repo_dir):
+                            # Not sure why the rename failed. Raise to user.
+                            raise
+
+            # If we get here, we have a bare repo with a remote set
+            # Pull content to update the cache. This should be safe for multiple doozer instances to perform.
+            self.logger.info(f'Updating cache directory for git remote: {remote_url}')
+            # Fire and forget this fetch -- just used to keep cache as fresh as possible
+            exectools.fire_and_forget(repo_dir, 'git fetch --all')
+            gitargs.extend(['--reference-if-able', repo_dir])
+
+        self.logger.info(f'Cloning to: {target_dir}')
+
+        # Perform the clone (including --reference args if cache_dir was set)
+        cmd = []
+        if timeout:
+            cmd.extend(['timeout', f'{timeout}'])
+        cmd.extend(['git', 'clone', remote_url])
+        cmd.extend(gitargs)
+        cmd.append(target_dir)
+        exectools.cmd_assert(cmd, retries=3, on_retry=["rm", "-rf", target_dir], set_env=set_env)
+
     def resolve_source(self, meta):
         """
         Looks up a source alias and returns a path to the directory containing
@@ -847,38 +906,38 @@ class Runtime(object):
 
         self.logger.debug("checking for source directory in source_dir: {}".format(source_dir))
 
-        # If this source has already been extracted for this working directory
-        if os.path.isdir(source_dir):
-            # Store so that the next attempt to resolve the source hits the map
-            self.source_paths[alias] = source_dir
-            self.logger.info("Source '{}' already exists in (skipping clone): {}".format(alias, source_dir))
-            return source_dir
+        with self.get_dir_lock(source_dir):
+            if alias in self.source_paths:  # we checked before, but check again inside the lock
+                self.logger.debug("returning previously resolved path for alias {}: {}".format(alias, self.source_paths[alias]))
+                return self.source_paths[alias]
 
-        clone_branch, _ = self.detect_remote_source_branch(source_details)
+            # If this source has already been extracted for this working directory
+            if os.path.isdir(source_dir):
+                # Store so that the next attempt to resolve the source hits the map
+                self.source_paths[alias] = source_dir
+                self.logger.info("Source '{}' already exists in (skipping clone): {}".format(alias, source_dir))
+                return source_dir
 
-        url = source_details["url"]
+            clone_branch, _ = self.detect_remote_source_branch(source_details)
+
+            url = source_details["url"]
 
-        self.logger.info("Attempting to checkout source '%s' branch %s in: %s" % (url, clone_branch, source_dir))
-        set_env = os.environ.copy()
-        set_env.update(constants.GIT_NO_PROMPTS)
-        try:
             self.logger.info("Attempting to checkout source '%s' branch %s in: %s" % (url, clone_branch, source_dir))
-            exectools.cmd_assert(
-                # get a little history to enable finding a recent Dockerfile change, but not too much.
+            try:
+                self.logger.info("Attempting to checkout source '%s' branch %s in: %s" % (url, clone_branch, source_dir))
                 # clone all branches as we must sometimes reference master /OWNERS for maintainer information
-                "git clone --no-single-branch -b {} {} --depth 50 {}".format(clone_branch, url, source_dir),
-                set_env=set_env,
-                retries=3,
-                on_retry=["rm", "-rf", source_dir],
-            )
+                gitargs = [
+                    '--no-single-branch', '--branch', clone_branch, '--depth', '1'
+                ]
+                self.git_clone(url, source_dir, gitargs=gitargs, set_env=constants.GIT_NO_PROMPTS)
 
-        except IOError as e:
-            self.logger.info("Unable to checkout branch {}: {}".format(clone_branch, str(e)))
-            raise DoozerFatalError("Error checking out target branch of source '%s' in: %s" % (alias, source_dir))
+            except IOError as e:
+                self.logger.info("Unable to checkout branch {}: {}".format(clone_branch, str(e)))
+                raise DoozerFatalError("Error checking out target branch of source '%s' in: %s" % (alias, source_dir))
 
-        # Store so that the next attempt to resolve the source hits the map
-        self.register_source_alias(alias, source_dir)
-        return source_dir
+            # Store so that the next attempt to resolve the source hits the map
+            self.register_source_alias(alias, source_dir)
+            return source_dir
 
     def detect_remote_source_branch(self, source_details):
         """Find a configured source branch that exists, or raise DoozerFatalError. Returns branch name and git hash"""
@@ -1013,12 +1072,13 @@ class Runtime(object):
         return ret.get(timeout)
 
     def clone_distgits(self, n_threads=None):
-        if n_threads is None:
-            n_threads = self.global_opts['distgit_threads']
-        return self._parallel_exec(
-            lambda m: m.distgit_repo(),
-            self.all_metas(),
-            n_threads=n_threads)
+        with util.timer(self.logger.info, 'Full runtime clone'):
+            if n_threads is None:
+                n_threads = self.global_opts['distgit_threads']
+            return self._parallel_exec(
+                lambda m: m.distgit_repo(),
+                self.all_metas(),
+                n_threads=n_threads)
 
     def push_distgits(self, n_threads=None):
         self.assert_mutation_is_permitted()
